﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Windows.Forms;
using System.ComponentModel;

// Some namespace mappings
using MF = MediaFoundation;

namespace Testing.SimpleMFPlayer.MediaPlayer
{
    public class MFMediaPlayer : Component, IDisposable
    {
        private const string CATEGORY_NAME = "Video Player";

        // Own properties
        private MFMediaPlayerState m_state;
        private Control m_targetControl;
        private string m_currentVideoFile;
        private TimeSpan m_currentVideoDuration;
        private bool m_isPaused;

        // References to Media Foundation
        private MF.IMFMediaSession m_mediaSession;
        private MF.EVR.IMFVideoDisplayControl m_displayControl;
        private MFSessionEventListener m_sessionEventHandler;

        /// <summary>
        /// Raised when the current state has changed.
        /// </summary>
        [Category(CATEGORY_NAME)]
        public event EventHandler StateChanged;

        /// <summary>
        /// Raises when the paused flag has changed.
        /// </summary>
        [Category(CATEGORY_NAME)]
        public event EventHandler IsPausedChanged;

        /// <summary>
        /// Initializes the <see cref="MFMediaPlayer"/> class.
        /// </summary>
        static MFMediaPlayer()
        {
            // Very important: Initialize MF library..
            MF.MFExtern.MFStartup(0x10070, MF.MFStartup.Full);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="MFMediaPlayer"/> class.
        /// </summary>
        public MFMediaPlayer()
        {
            m_currentVideoFile = string.Empty;
        }

        /// <summary>
        /// Pauses the video currently playing.
        /// </summary>
        public async Task PauseVideoAsync()
        {
            // Check for correct state
            if (this.State != MFMediaPlayerState.Playing)
            {
                throw new InvalidOperationException("Unable to pause video as long as there is no video playing!");
            }

            // Video already paused?
            if (m_isPaused) { return; }

            // Pause the video
            Task waitTask = m_sessionEventHandler.WaitForEventAsync(MF.MediaEventType.MESessionPaused, CancellationToken.None);
            m_mediaSession.Pause();
            await waitTask;

            // Updates current paused state
            this.IsPaused = true;
        }

        /// <summary>
        /// Resumes the video from current position.
        /// </summary>
        public async Task ResumeVideoAsync()
        {
            // Check for correct state
            if (this.State != MFMediaPlayerState.Playing)
            {
                throw new InvalidOperationException("Unable to resume video as long as there is no video playing!");
            }

            // Video is not paused?
            if (!m_isPaused) { return; }

            // Resume the video from current location
            await StartSessionInternalAsync(false);

            // Updates current paused state
            this.IsPaused = false;
        }

        /// <summary>
        /// Opens the given video file and plays it directly.
        /// </summary>
        /// <param name="sourceVideoFile">The source video file.</param>
        public async Task OpenAndShowVideoFileAsync(string sourceVideoFile)
        {
            // Check for correct state
            if (this.State != MFMediaPlayerState.NothingToDo)
            {
                throw new InvalidOperationException("Unable to open video file as long as there is another video playing!");
            }

            // Apply new state
            this.State = MFMediaPlayerState.Opening;

            try
            {
                // Create media session and a corresponding event listener obect for async events
                MF.MFExtern.MFCreateMediaSession(null, out m_mediaSession);
                m_sessionEventHandler = MFSessionEventListener.AttachTo(m_mediaSession);
                m_sessionEventHandler.EndOfPresentation += OnSessionEventHandlerEndOfPresentationReached;

                // Create source object
                MF.IMFSourceResolver sourceResolver;
                MF.MFExtern.MFCreateSourceResolver(out sourceResolver);
                MF.MFObjectType objType = MF.MFObjectType.Invalid;
                object objSource = null;
                sourceResolver.CreateObjectFromURL(
                    sourceVideoFile,
                    MF.MFResolution.MediaSource,
                    null,
                    out objType,
                    out objSource);
                MF.IMFMediaSource mediaSource = objSource as MF.IMFMediaSource;

                // Create topology
                MF.IMFTopology topology;
                MF.IMFPresentationDescriptor presentationDescriptor;
                MF.MFExtern.MFCreateTopology(out topology);
                mediaSource.CreatePresentationDescriptor(out presentationDescriptor);
                int streamDescriptorCount = 0;
                presentationDescriptor.GetStreamDescriptorCount(out streamDescriptorCount);
                for (int loop = 0; loop < streamDescriptorCount; loop++)
                {
                    bool selected = false;
                    MF.IMFStreamDescriptor streamDescriptor;
                    presentationDescriptor.GetStreamDescriptorByIndex(loop, out selected, out streamDescriptor);
                    if (selected)
                    {
                        // Create source node
                        MF.IMFTopologyNode sourceNode = null;
                        MF.MFExtern.MFCreateTopologyNode(MF.MFTopologyType.SourcestreamNode, out sourceNode);
                        sourceNode.SetUnknown(MF.MFAttributesClsid.MF_TOPONODE_SOURCE, mediaSource);
                        sourceNode.SetUnknown(MF.MFAttributesClsid.MF_TOPONODE_PRESENTATION_DESCRIPTOR, presentationDescriptor);
                        sourceNode.SetUnknown(MF.MFAttributesClsid.MF_TOPONODE_STREAM_DESCRIPTOR, streamDescriptor);

                        // Create output node
                        MF.IMFTopologyNode outputNode = null;
                        MF.IMFMediaTypeHandler mediaTypeHandler = null;
                        streamDescriptor.GetMediaTypeHandler(out mediaTypeHandler);
                        Guid majorType = Guid.Empty;
                        mediaTypeHandler.GetMajorType(out majorType);
                        MF.MFExtern.MFCreateTopologyNode(MF.MFTopologyType.OutputNode, out outputNode);
                        if (MF.MFMediaType.Audio == majorType)
                        {
                            MF.IMFActivate audioRenderer;
                            MF.MFExtern.MFCreateAudioRendererActivate(out audioRenderer);
                            outputNode.SetObject(audioRenderer);
                        }
                        else if (MF.MFMediaType.Video == majorType)
                        {
                            MF.IMFActivate videoRenderer;
                            MF.MFExtern.MFCreateVideoRendererActivate(
                                m_targetControl.Handle,
                                out videoRenderer);
                            outputNode.SetObject(videoRenderer);
                        }

                        // Append nodes to topology
                        topology.AddNode(sourceNode);
                        topology.AddNode(outputNode);
                        sourceNode.ConnectOutput(0, outputNode, 0);
                    }
                }

                // Get the total duration of the video
                long durationLong = 0;
                presentationDescriptor.GetUINT64(MF.MFAttributesClsid.MF_PD_DURATION, out durationLong);
                m_currentVideoDuration = TimeSpan.FromTicks(durationLong);

                // Apply build topology to the session 
                Task<MF.IMFMediaEvent> topologyReadyWaiter = m_sessionEventHandler.WaitForEventAsync(
                    MF.MediaEventType.MESessionTopologyStatus,
                    (eventData) => 
                    {
                        int topologyState = 0;
                        eventData.GetUINT32(MF.MFAttributesClsid.MF_EVENT_TOPOLOGY_STATUS, out topologyState);
                        return topologyState == (int)(MF.MFTopoStatus.Ready);
                    },
                    CancellationToken.None);
                m_mediaSession.SetTopology(MF.MFSessionSetTopologyFlags.None, topology);
                await topologyReadyWaiter;

                // Query for display control service
                object displayControlObject = null;
                MF.MFExtern.MFGetService(
                    m_mediaSession,
                    MF.MFServices.MR_VIDEO_RENDER_SERVICE,
                    typeof(MF.EVR.IMFVideoDisplayControl).GUID,
                    out displayControlObject);
                m_displayControl = displayControlObject as MF.EVR.IMFVideoDisplayControl;

                // Start playing the video
                await StartSessionInternalAsync(true);

                // Video opened successfully
                m_currentVideoFile = sourceVideoFile;
                this.State = MFMediaPlayerState.Playing;
            }
            catch (Exception)
            {
                // Unload all resources in case of an exception
                DisposeResources();

                throw;
            }
        }

        /// <summary>
        /// Starts playing the video.
        /// </summary>
        private async Task StartSessionInternalAsync(bool restart)
        {
            MF.Misc.PropVariant startTimeStamp;

            // Build play parameters
            if (restart)
            {
                startTimeStamp = new MF.Misc.PropVariant((long)0);
                //startTimeStamp.Type = SharpDX.Win32.VariantType.Default;
                //startTimeStamp.ElementType = SharpDX.Win32.VariantElementType.Long;
                //startTimeStamp.Value = (long)0;
            }
            else
            {
                startTimeStamp = new MF.Misc.PropVariant();
                //startTimeStamp.Type = SharpDX.Win32.VariantType.Default;
                //startTimeStamp.ElementType = SharpDX.Win32.VariantElementType.Empty;
            }

            // Start playing the video
            Task waitTask = m_sessionEventHandler.WaitForEventAsync(MF.MediaEventType.MESessionStarted, CancellationToken.None);
            m_mediaSession.Start(Guid.Empty, startTimeStamp);
            await waitTask;
        }

        /// <summary>
        /// Closes the current video.
        /// This method returns immediately if there is no video playing.
        /// </summary>
        public async Task CloseVideoAsync()
        {
            // Handle current state
            if (this.State != MFMediaPlayerState.Playing) { return; }
            this.State = MFMediaPlayerState.Closing;

            try
            {
                // Call close and wait for result
                Task<MF.IMFMediaEvent> closeSessionTask = m_sessionEventHandler.WaitForEventAsync(
                    MF.MediaEventType.MESessionClosed,
                    CancellationToken.None);
                m_mediaSession.Close();
                await closeSessionTask;
            }
            finally
            {
                // Disposes all resources
                this.DisposeResources();
            }
        }

        /// <summary>
        /// Current session reports finished video.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private async void OnSessionEventHandlerEndOfPresentationReached(object sender, EventArgs e)
        {
            if (sender != m_sessionEventHandler) { return; }

            if (this.RestartVideoWhenFinished)
            {
                // Start the session again
                await StartSessionInternalAsync(true);
            }
            else
            {
                // Close current video
                await CloseVideoAsync();
            }
        }

        /// <summary>
        /// Called when the target control changes its size.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void OnTargetControlResize(object sender, EventArgs e)
        {
            if((m_displayControl != null) &&
               (m_targetControl != null))
            {
                m_displayControl.SetVideoPosition(
                    null,
                    new MF.Misc.MFRect(0, 0, m_targetControl.Width, m_targetControl.Height));
            }
        }

        /// <summary>
        /// Called when the target control requested painting.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="PaintEventArgs"/> instance containing the event data.</param>
        private void OnTargetControlPaint(object sender, PaintEventArgs e)
        {
            if ((m_displayControl != null) &&
                (m_targetControl != null))
            {
                m_displayControl.RepaintVideo();
            }
        }

        /// <summary>
        /// Disposes all current resources.
        /// </summary>
        private void DisposeResources()
        {
            // Shutdown current session
            if (m_mediaSession != null)
            {
                m_mediaSession.Shutdown();
            }

            // Clear all references
            m_currentVideoFile = string.Empty;
            m_currentVideoDuration = TimeSpan.Zero;
            m_sessionEventHandler = null;
            m_displayControl = null;

            // Apply new state
            this.IsPaused = false;
            this.State = MFMediaPlayerState.NothingToDo;
        }

        /// <summary>
        /// Gets the state of the media player.
        /// </summary>
        [Browsable(false)]
        public MFMediaPlayerState State
        {
            get { return m_state; }
            private set
            {
                if(m_state != value)
                {
                    m_state = value;
                    if (StateChanged != null) { StateChanged(this, EventArgs.Empty); }
                }
            }
        }

        /// <summary>
        /// Gets a reference to the current displayed video.
        /// </summary>
        [Browsable(false)]
        public string CurrentVideoFile
        {
            get { return m_currentVideoFile; }
        }

        /// <summary>
        /// Gets the total duration of the currently displayed video.
        /// </summary>
        [Browsable(false)]
        public TimeSpan CurrentVideoDuration
        {
            get { return m_currentVideoDuration; }
        }

        /// <summary>
        /// Restart the video when playing has finished?
        /// </summary>
        [Category(CATEGORY_NAME)]
        public bool RestartVideoWhenFinished
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the target control which is used for displaying the video.
        /// </summary>
        [Category(CATEGORY_NAME)]
        public Control TargetControl
        {
            get { return m_targetControl; }
            set
            {
                // Check current state
                if (this.State != MFMediaPlayerState.NothingToDo)
                {
                    throw new InvalidOperationException("Unable to change TargetControl property as long there is a video playing!");
                }

                // Switch the target control
                if(m_targetControl != value)
                {
                    if(m_targetControl != null)
                    {
                        m_targetControl.Resize -= OnTargetControlResize;
                        m_targetControl.Paint -= OnTargetControlPaint;
                    }
                    m_targetControl = value;
                    if(m_targetControl != null)
                    {
                        m_targetControl.Resize += OnTargetControlResize;
                        m_targetControl.Paint += OnTargetControlPaint;
                    }
                }
            }
        }

        /// <summary>
        /// Is the video paused currently?
        /// </summary>
        [Browsable(false)]
        public bool IsPaused 
        {
            get { return m_isPaused; }
            private set
            {
                if(m_isPaused != value)
                {
                    m_isPaused = value;
                    if (IsPausedChanged != null) { IsPausedChanged(this, EventArgs.Empty); }
                }
            }
        }
    }
}
