﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

// Some namespace mappings
using MF = MediaFoundation;

namespace Testing.SimpleMFPlayer.MediaPlayer
{
    class MFSessionEventListener : MF.Misc.COMBase, MF.IMFAsyncCallback
    {
        // The captured synchronization context
        private SynchronizationContext m_syncContext;

        // Helper for WaitForEventAsync method
        private List<Func<MF.IMFMediaEvent, bool>> m_asyncEventWaiters;
        private object m_asyncEventWaitersLock;

        // The session this event listener listens to
        private MF.IMFMediaSession m_session;

        /// <summary>
        /// Raises when the end of a video is reached.
        /// </summary>
        public event EventHandler EndOfPresentation;

        /// <summary>
        /// Initializes a new instance of the <see cref="MFSessionEventListener"/> class.
        /// </summary>
        /// <param name="session">The session to listen to.</param>
        private MFSessionEventListener(MF.IMFMediaSession session)
        {
            m_syncContext = SynchronizationContext.Current;
            m_session = session;

            m_asyncEventWaiters = new List<Func<MF.IMFMediaEvent, bool>>();
            m_asyncEventWaitersLock = new object();

            // Start listening on the session object
            m_session.BeginGetEvent(this, null);
        }

        /// <summary>
        /// Attaches this listener on the given session.
        /// </summary>
        /// <param name="session">The session object to attach to.</param>
        internal static MFSessionEventListener AttachTo(MF.IMFMediaSession session)
        {
            return new MFSessionEventListener(session);
        }

        /// <summary>
        /// This method waits for an asynchronous event message from the MF framework.
        /// </summary>
        /// <param name="eventType">The type of the event to wait for.</param>
        /// <param name="specificEventHandler">A specific event handler implemented by the caller. Returns false if the event is not handled.</param>
        /// <param name="cancelToken">The cancellation token.</param>
        internal Task<MF.IMFMediaEvent> WaitForEventAsync(
            MF.MediaEventType eventType, 
            CancellationToken cancelToken)
        {
            return WaitForEventAsync(eventType, null, cancelToken);
        }

        /// <summary>
        /// This method waits for an asynchronous event message from the MF framework.
        /// </summary>
        /// <param name="eventType">The type of the event to wait for.</param>
        /// <param name="specificEventCheckingFunc">A specific event checking func implemented by the caller. Returns false if the event is not handled.</param>
        /// <param name="cancelToken">The cancellation token.</param>
        internal Task<MF.IMFMediaEvent> WaitForEventAsync(
            MF.MediaEventType eventType, 
            Func<MF.IMFMediaEvent, bool> specificEventCheckingFunc,
            CancellationToken cancelToken)
        {
            TaskCompletionSource<MF.IMFMediaEvent> eventWaitTaskResult = new TaskCompletionSource<MF.IMFMediaEvent>();

            // Define the wait action
            Func<MF.IMFMediaEvent, bool> waitFunc = (givenEventData) =>
            {
                // Handle cancellation
                if (cancelToken.IsCancellationRequested)
                {
                    eventWaitTaskResult.TrySetCanceled();
                    return true;
                }

                // Is this event that one we are waiting for?
                MF.MediaEventType givenEventType;
                givenEventData.GetType(out givenEventType);
                if (givenEventType != eventType)
                {
                    return false;
                }

                // Handle result of the given event
                int eventStatus = 0;
                givenEventData.GetStatus(out eventStatus);
                if (!Succeeded(eventStatus))
                {
                    eventWaitTaskResult.TrySetException(new ApplicationException("MF reported HRESULT " + eventStatus + " for event " + eventType + "!"));
                }
                else
                {
                    // If given, then execute special event checking
                    if (specificEventCheckingFunc != null)
                    {
                        if (!specificEventCheckingFunc(givenEventData)) { return false; }
                    }

                    // Finish task 
                    eventWaitTaskResult.TrySetResult(givenEventData);
                }
                return true;
            };

            // Register waiter
            lock(m_asyncEventWaitersLock)
            {
                m_asyncEventWaiters.Add(waitFunc);
            }

            return eventWaitTaskResult.Task;
        }

        /// <summary>
        ///   <p><strong>Applies to: </strong>desktop apps | Metro style apps</p><p> </p><p>Called when an asynchronous operation is completed.</p>
        /// </summary>
        /// <param name="asyncResultRef"><dd> <p>Pointer to the <strong><see cref="T:SharpDX.MediaFoundation.AsyncResult" /></strong> interface. Pass this reference to the asynchronous <strong>End...</strong> method to complete the asynchronous call.</p> </dd></param>
        /// <!-- Failed to insert some or all of included XML --><include file=".\..\Documentation\CodeComments.xml" path="/comments/comment[@id='IMFAsyncCallback::Invoke']/*" />
        ///   <msdn-id>bb970360</msdn-id>
        ///   <unmanaged>HRESULT IMFAsyncCallback::Invoke([In, Optional] IMFAsyncResult* pAsyncResult)</unmanaged>
        ///   <unmanaged-short>IMFAsyncCallback::Invoke</unmanaged-short>
        /// <remarks>
        ///   <p>Within your implementation of <strong>Invoke</strong>, call the corresponding <strong>End...</strong> method.</p><p>This interface is available on the following platforms if the Windows Media Format 11 SDK redistributable components are installed:</p><ul> <li>Windows?XP with Service Pack?2 (SP2) and later.</li> <li>Windows?XP Media Center Edition?2005 with KB900325 (Windows?XP Media Center Edition?2005) and KB925766 (October 2006 Update Rollup for Windows?XP Media Center Edition) installed.</li> </ul>
        /// </remarks>
        public void Invoke(MF.IMFAsyncResult asyncResultRef)
        {
            MF.MediaEventType eventType;
            MF.IMFMediaEvent mediaEvent = null;

            try
            {
                m_session.EndGetEvent(asyncResultRef, out mediaEvent);

                // Look for existing awaitor action and handle it if there is one
                lock (m_asyncEventWaitersLock)
                {
                    for (int loop = 0; loop < m_asyncEventWaiters.Count; loop++)
                    {
                        if (m_asyncEventWaiters[loop](mediaEvent))
                        {
                            m_asyncEventWaiters.RemoveAt(loop);
                            break;
                        }
                    }
                }

                // Some generic event handling
                mediaEvent.GetType(out eventType);
                switch (eventType)
                {
                    case MF.MediaEventType.MEEndOfPresentation:
                        m_syncContext.Post((arg) =>
                        {
                            if (EndOfPresentation != null) { EndOfPresentation(this, EventArgs.Empty); }
                        }, null);
                        break;
                }
            }
            catch (Exception ex)
            {
                // Dispatch exception to UI thread
                m_syncContext.Post((arg) =>
                {
                    Application.OnThreadException(ex);
                }, null);
            }

            // Reregister for next event
            m_session.BeginGetEvent(this, null); 
        }

        /// <summary>
        /// Gets the parameters.
        /// </summary>
        /// <param name="pdwFlags">The PDW flags.</param>
        /// <param name="pdwQueue">The PDW queue.</param>
        public void GetParameters(out MF.MFASync pdwFlags, out MF.MFAsyncCallbackQueue pdwQueue)
        {
            pdwFlags = MF.MFASync.FastIOProcessingCallback;
            pdwQueue = MF.MFAsyncCallbackQueue.Standard;
        }
    }
}
